package org.processmining.models.graphbased.directed.socialnetwork;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.SwingConstants;

import org.processmining.framework.providedobjects.SubstitutionType;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.AbstractDirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;

// objects of this type should be represented in the framework by the
// SocialNetwork interface.
@SubstitutionType(substitutedType = SocialNetwork.class)
public class SocialNetworkImpl extends AbstractDirectedGraph<SNNode, SNEdge> implements SocialNetwork {

	private final Set<SNNode> SNNodes = new LinkedHashSet<SNNode>();
	private final Set<SNEdge> SNEdges = new LinkedHashSet<SNEdge>();;

	public SocialNetworkImpl(String label) {
		super();
		getAttributeMap().put(AttributeMap.LABEL, label);
		getAttributeMap().put(AttributeMap.PREF_ORIENTATION, SwingConstants.NORTH);
	}

	@Override
	protected synchronized Map<DirectedGraphElement, DirectedGraphElement> cloneFrom(DirectedGraph<SNNode, SNEdge> graph) {
		assert (graph instanceof SocialNetworkImpl);
		Map<DirectedGraphElement, DirectedGraphElement> mapping = new HashMap<DirectedGraphElement, DirectedGraphElement>();

		SocialNetworkImpl ts = (SocialNetworkImpl) graph;
		for (SNNode SNNode : ts.SNNodes) {
			addSNNode(SNNode.getIdentifier());
			mapping.put(SNNode, findSNNode(SNNode.getIdentifier()));
		}
		for (SNEdge trans : ts.SNEdges) {
			addSNEdge(trans.getSource(), trans.getTarget(), trans.getIdentifier());
			mapping.put(trans, findSNEdge(trans.getSource().getIdentifier(), trans.getTarget().getIdentifier(), trans
					.getIdentifier()));
		}
		return mapping;
	}

	@Override
	protected synchronized AbstractDirectedGraph<SNNode, SNEdge> getEmptyClone() {
		return new SocialNetworkImpl(getLabel());
	}

	@SuppressWarnings("unchecked")
	@Override
	public synchronized void removeEdge(DirectedGraphEdge edge) {
		assert (edge instanceof SNEdge);
		SNEdge t = (SNEdge) edge;
		removeSNEdge(t);
	}

	public synchronized void removeNode(DirectedGraphNode node) {
		if (node instanceof SNNode) {
			removeSNNodeInternal((SNNode) node);
		} else {
			assert (false);
		}
	}

	public synchronized Collection<SNEdge> getSNEdges() {
		Collection<SNEdge> edges = new ArrayList<SNEdge>(SNEdges.size());
		for (SNEdge edge : SNEdges) {
			edges.add(edge);
		}
		return edges;
	}

	private synchronized Object removeSNEdge(SNEdge SNEdge) {
		SNEdge result = removeNodeFromCollection(SNEdges, SNEdge);
		return (result == null ? null : result.getIdentifier());
	}

	public synchronized Set<SNEdge> getEdges() {
		return SNEdges;
	}

	public double getMaxFlowValue() {
		double maxValue = Double.MIN_VALUE;
		for (SNEdge edge : SNEdges) {
			if (edge.getWeight() > maxValue) {
				maxValue = edge.getWeight();
			}
		}
		return maxValue;
	}

	public double getMinFlowValue() {
		double minValue = Double.MAX_VALUE;
		for (SNEdge edge : SNEdges) {
			if (edge.getWeight() < minValue) {
				minValue = edge.getWeight();
			}
		}
		return minValue;
	}

	public synchronized Set<SNNode> getNodes() {
		return SNNodes;
	}

	public synchronized boolean addSNNode(Object identifier) {
		SNNode snNode = new SNNode(identifier, this);
		if (SNNodes.add(snNode)) {
			graphElementAdded(snNode);
			return true;
		} else {
			return false;
		}
	}

	public synchronized boolean addSNEdge(Object fromNodeID, Object toNodeID, Object identifier) {
		SNNode source = findSNNode(fromNodeID);
		SNNode target = findSNNode(toNodeID);
		checkAddEdge(source, target);
		SNEdge trans = new SNEdge(source, target, identifier);
		if (SNEdges.add(trans)) {
			graphElementAdded(trans);
			return true;
		} else {
			return false;
		}
	}

	public synchronized boolean addSNEdge(SNNode fromSNNode, SNNode toSNNode, Object identifier) {
		checkAddEdge(fromSNNode, toSNNode);
		SNEdge trans = new SNEdge(fromSNNode, toSNNode, identifier);
		if (SNEdges.add(trans)) {
			graphElementAdded(trans);
			return true;
		} else {
			return false;
		}
	}

	public synchronized Collection<SNNode> getSNNodes() {
		Collection<SNNode> stat = new ArrayList<SNNode>(SNNodes.size());
		for (SNNode tsSNNode : SNNodes) {
			stat.add(tsSNNode);
		}
		return stat;
	}

	public synchronized Object removeSNNode(Object identifier) {
		return removeSNNodeInternal(findSNNode(identifier));
	}

	private synchronized Object removeSNNodeInternal(SNNode SNNode) {
		SNNode result = removeNodeFromCollection(SNNodes, findSNNode(SNNode));
		return (result == null ? null : result.getIdentifier());
	}

	public synchronized SNNode findSNNode(Object identifier) {
		for (SNNode s : SNNodes) {
			if (s.getIdentifier().equals(identifier)) {
				return s;
			}
		}
		return null;
	}

	public synchronized SNEdge findSNEdge(Object fromSNNode, Object toSNNode, Object identifier) {
		for (SNEdge t : getEdges(findSNNode(fromSNNode), findSNNode(toSNNode), getEdges())) {
			if (t.getIdentifier().equals(identifier)) {
				return t;
			}
		}
		return null;
	}

	public synchronized Object removeSNEdge(SNNode fromSNNode, SNNode toSNNode, Object identifier) {
		return removeSNEdge(new SNEdge(findSNNode(fromSNNode.getIdentifier()), findSNNode(toSNNode.getIdentifier()),
				identifier));
	}

	// additional 
	public int getInDegree(SNNode snNode) {
		int degree = 0;
		for (SNEdge edge : SNEdges) {
			if (edge.getTarget().equals(snNode)) {
				degree++;
			}
		}
		return degree;
	}

	public int getOutDegree(SNNode snNode) {
		int degree = 0;
		for (SNEdge edge : SNEdges) {
			if (edge.getSource().equals(snNode)) {
				degree++;
			}
		}
		return degree;
	}

	public int getDegree(SNNode snNode) {
		int degree = 0;
		for (SNEdge edge : SNEdges) {
			if (edge.getSource().equals(snNode) || edge.getTarget().equals(snNode)) {
				degree++;
			}
		}
		return degree;
	}

	public double getInWeightDegree(SNNode snNode) {
		double degree = 0.0;
		for (SNEdge edge : SNEdges) {
			if (edge.getTarget().equals(snNode)) {
				degree += edge.getWeight();
			}
		}
		return degree;
	}

	public double getOutWeightDegree(SNNode snNode) {
		double degree = 0;
		for (SNEdge edge : SNEdges) {
			if (edge.getSource().equals(snNode)) {
				degree += edge.getWeight();
			}
		}
		return degree;
	}

	public double getWeightDegree(SNNode snNode) {
		double degree = 0;
		for (SNEdge edge : SNEdges) {
			if (edge.getSource().equals(snNode) || edge.getTarget().equals(snNode)) {
				degree += edge.getWeight();
			}
		}
		return degree;
	}
}
