package org.processmining.socialnetwork.models.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.ArrayUtils;
import org.processmining.framework.util.Pair;
import org.processmining.socialnetwork.models.SocialNetwork;
import org.processmining.socialnetwork.models.SocialNetworkArcLabelStyle;
import org.processmining.socialnetwork.models.SocialNetworkArcStyle;
import org.processmining.socialnetwork.models.SocialNetworkVertexStyle;
import org.processmining.socialnetwork.models.SocialNetworkVisualization;
import org.processmining.socialnetwork.models.SocialNetworkVisualizationAlgorithm;

import gnu.trove.iterator.TObjectDoubleIterator;
import gnu.trove.map.TObjectDoubleMap;
import gnu.trove.map.hash.TObjectDoubleHashMap;
import gnu.trove.set.hash.THashSet;

public class WindowedSocialNetworkVisualizationAlgorithmImpl<T> extends AbstractSocialNetworkVisualizationAlgorithm<T>{

	private final static String PARAMETER_KEY_WINDOW_SIZE = "window_size";

	private TObjectDoubleMap<Pair<T, T>> avgWindow1 = null;

	//TODO: Make parameter
	private double threshold = 0.0;

	private List<SocialNetwork<T>> window1 = new ArrayList<>();
	private Collection<Pair<T, T>> window1Deleted = new THashSet<>();
	private List<SocialNetwork<T>> window2 = new ArrayList<>();
	private int windowSize = 50;

	private Object monitor = new Object();

	public WindowedSocialNetworkVisualizationAlgorithmImpl() {
		super();
	}
	
	public WindowedSocialNetworkVisualizationAlgorithmImpl(SocialNetworkVisualization<T> visualization) {
		super(visualization);
	}
	

	public void clear() {
		synchronized (monitor) {
			avgWindow1.clear();
			window1.clear();
			window2.clear();
			window1Deleted.clear();
			getVisualization().clear();
		}

	}

	private TObjectDoubleMap<Pair<T, T>> computeAverage(final List<SocialNetwork<T>> window) {
		TObjectDoubleMap<Pair<T, T>> average = new TObjectDoubleHashMap<>();
		for (SocialNetwork<T> nw : window) {
			for (T from : nw.getVertices()) {
				for (T to : nw.getOutSet(from)) {
					double arcValue = nw.getArcValue(from, to);
					average.adjustOrPutValue(new Pair<T, T>(from, to), arcValue, arcValue);
				}
			}
		}
		// divide
		int divisor = window.size();
		for (TObjectDoubleIterator<Pair<T, T>> it = average.iterator(); it.hasNext();) {
			it.advance();
			it.setValue(it.value() / divisor);
		}
		return average;
	}

	private TObjectDoubleMap<T> computeNormalizedResourceValues(TObjectDoubleMap<Pair<T, T>> avgWindow2,
			TObjectDoubleMap<Pair<T, T>> relativeChanges) {
		TObjectDoubleMap<T> relativeResValues = new TObjectDoubleHashMap<>();
		TObjectDoubleMap<T> absoluteResValues = new TObjectDoubleHashMap<>();
		for (TObjectDoubleIterator<Pair<T, T>> it = relativeChanges.iterator(); it.hasNext();) {
			it.advance();
			if (it.value() == Double.MIN_VALUE) {
				absoluteResValues.adjustOrPutValue(it.key().getFirst(), 0, 0);
			} else {
				double edgeVal = avgWindow2.get(it.key());
				absoluteResValues.adjustOrPutValue(it.key().getFirst(), edgeVal, edgeVal);
				absoluteResValues.adjustOrPutValue(it.key().getSecond(), edgeVal, edgeVal);
			}
		}
		Collection<Double> absoluteValues = Arrays.asList(ArrayUtils.toObject(absoluteResValues.values()));
		double max = Collections.max(absoluteValues);
		double min = Collections.min(absoluteValues);
		double div = max - min;
		for (TObjectDoubleIterator<T> it = absoluteResValues.iterator(); it.hasNext();) {
			it.advance();
			double relative = (it.value() - min) / div;
			relativeResValues.put(it.key(), relative);
		}
		return relativeResValues;
	}

	private TObjectDoubleMap<Pair<T, T>> computeRelativeChanges(TObjectDoubleMap<Pair<T, T>> avgWindow1,
			TObjectDoubleMap<Pair<T, T>> avgWindow2) {
		TObjectDoubleMap<Pair<T, T>> relativeChanges = new TObjectDoubleHashMap<>();
		for (TObjectDoubleIterator<Pair<T, T>> it = avgWindow2.iterator(); it.hasNext();) {
			it.advance();
			if (avgWindow1.containsKey(it.key())) {
				double val = (it.value() / avgWindow1.get(it.key())) - 1;
				relativeChanges.put(it.key(), val);
			} else {
				relativeChanges.put(it.key(), Double.MAX_VALUE);
			}
		}
		Set<Pair<T, T>> leavers = new THashSet<>(avgWindow1.keySet());
		leavers.removeAll(avgWindow2.keySet());
		for (Pair<T, T> leaver : leavers) {
			relativeChanges.put(leaver, Double.MIN_VALUE);
		}
		return relativeChanges;
	}

	private SocialNetworkArcStyle getEdgeStyleBasedOnNormalizedValue(double val) {
		if (val < 0.2) {
			return SocialNetworkArcStyle.VERY_LOW;
		} else if (val < 0.4) {
			return SocialNetworkArcStyle.LOW;
		} else if (val < 0.6) {
			return SocialNetworkArcStyle.NEUTRAL;
		} else if (val < 0.8) {
			return SocialNetworkArcStyle.HIGH;
		} else {
			return SocialNetworkArcStyle.VERY_HIGH;
		}
	}

	private SocialNetworkVertexStyle getNodeStyleBasedOnNormalizedValue(double val) {
		if (val < 0.2) {
			return SocialNetworkVertexStyle.VERY_LOW;
		} else if (val < 0.4) {
			return SocialNetworkVertexStyle.LOW;
		} else if (val < 0.6) {
			return SocialNetworkVertexStyle.NEUTRAL;
		} else if (val < 0.8) {
			return SocialNetworkVertexStyle.HIGH;
		} else {
			return SocialNetworkVertexStyle.VERY_HIGH;
		}
	}

	@Override
	public List<String> getParameterValues() {
		List<String> values = super.getParameterValues();
		values.add(Integer.toString(windowSize));
		return values;
	}

	private <U> TObjectDoubleMap<U> normalize(final TObjectDoubleMap<U> input, final TObjectDoubleMap<U> normalized) {
		Collection<Double> absoluteValues = Arrays.asList(ArrayUtils.toObject(input.values()));
		double max = Collections.max(absoluteValues);
		double min = Collections.min(absoluteValues);
		double div = max - min;
		for (TObjectDoubleIterator<U> it = input.iterator(); it.hasNext();) {
			it.advance();
			double value = (it.value() - min) / div;
			normalized.put(it.key(), value);
		}
		return normalized;
	}

	public void processNewNetwork(SocialNetwork<T> network) {
		synchronized (monitor) {
			updateWindows(network);
		}
	}

	@Override
	public void setParameter(String key, String value) {
		super.setParameter(key, value);
		if (key.equals(PARAMETER_KEY_WINDOW_SIZE)) {
			this.windowSize = Integer.valueOf(value);
		}
	}

	protected void setupParameterKeyList(List<String> list) {
		list.add(PARAMETER_KEY_WINDOW_SIZE);
	}

	private void shiftWindows() {
		if (avgWindow1 == null) {
			avgWindow1 = computeAverage(window1);
		}
		TObjectDoubleMap<Pair<T, T>> avgWindow2 = computeAverage(window2);
		TObjectDoubleMap<Pair<T, T>> relativeChanges = computeRelativeChanges(avgWindow1, avgWindow2);
		updateNetworkWithNewWindow(avgWindow2, relativeChanges);
		window1Deleted.removeAll(avgWindow2.keySet());
		for (Pair<T, T> remove : window1Deleted) {
			getVisualization().removeArc(remove.getFirst(), remove.getSecond());
		}
		window1Deleted.clear();
		for (TObjectDoubleIterator<Pair<T, T>> it = relativeChanges.iterator(); it.hasNext();) {
			it.advance();
			if (it.value() == Double.MIN_VALUE) {
				window1Deleted.add(it.key());
			}
		}
		window1 = window2;
		window2 = new ArrayList<SocialNetwork<T>>();
		avgWindow1 = avgWindow2;
	}

	private void updateNetworkWithNewWindow(TObjectDoubleMap<Pair<T, T>> avgWindow2,
			TObjectDoubleMap<Pair<T, T>> relativeChanges) {
		TObjectDoubleMap<Pair<T, T>> normalizedAvgW2 = normalize(avgWindow2, new TObjectDoubleHashMap<Pair<T, T>>());
		// decorate edges...
		for (TObjectDoubleIterator<Pair<T, T>> it = relativeChanges.iterator(); it.hasNext();) {
			it.advance();
			T from = it.key().getFirst();
			T to = it.key().getSecond();
			if (!getVisualization().containsArc(from, to)) {
				if (!getVisualization().containsVertex(from)) {
					getVisualization().addVertex(from);
					getVisualization().setVertexLabel(from, from.toString());
				}
				if (!getVisualization().containsVertex(to)) {
					getVisualization().addVertex(to);
					getVisualization().setVertexLabel(to, to.toString());
				}
				getVisualization().addArc(from, to);
			}
			if (avgWindow2.get(it.key()) > threshold || it.value() == Double.MIN_VALUE) {
				//				SocialNetworkArc<String> link = getOrAddLink(it.key());
				if (it.value() == Double.MIN_VALUE) {
					getVisualization().setArcLabel(from, to, "X");
					getVisualization().setArcStyle(from, to, SocialNetworkArcStyle.DELETED);
					getVisualization().setArcLabelStyle(from, to, SocialNetworkArcLabelStyle.DOWN);
				} else {
					double normalizedLinkValue = normalizedAvgW2.get(it.key());
					if (it.value() == Double.MAX_VALUE) {
						getVisualization().setArcStyle(from, to,
								getEdgeStyleBasedOnNormalizedValue(normalizedLinkValue));
						getVisualization().setArcLabel(from, to, "new");
						getVisualization().setArcLabelStyle(from, to, SocialNetworkArcLabelStyle.UP);
					} else {
						getVisualization().setArcLabel(from, to, Double.toString(avgWindow2.get(it.key())));
						getVisualization().setArcStyle(from, to,
								getEdgeStyleBasedOnNormalizedValue(normalizedLinkValue));
						boolean up = it.value() > 0;
						String label = up ? "\u2191\t\t\t" : "\u2193\t\t\t";
						getVisualization().setArcLabel(from, to, label);
						if (up) {
							getVisualization().setArcLabelStyle(from, to, SocialNetworkArcLabelStyle.UP);
						} else {
							getVisualization().setArcLabelStyle(from, to, SocialNetworkArcLabelStyle.DOWN);
						}
					}
				}
			} else if (avgWindow1.containsKey(it.key())) {
				// no longer above threshold
				getVisualization().removeArc(from, to);
			}
		}
		TObjectDoubleMap<T> resourceNormalized = computeNormalizedResourceValues(avgWindow2, relativeChanges);
		for (TObjectDoubleIterator<T> it = resourceNormalized.iterator(); it.hasNext();) {
			it.advance();
			getVisualization().setVertexStyle(it.key(), getNodeStyleBasedOnNormalizedValue(it.value()));
		}
	}

	private void updateWindows(final SocialNetwork<T> newNetwork) {
		if (window1.size() < windowSize) {
			window1.add(newNetwork);
		} else if (window2.size() < windowSize) {
			window2.add(newNetwork);
		} else if (window2.size() % windowSize == 0) {
			shiftWindows();
		}
	}

	public SocialNetworkVisualizationAlgorithm.Algorithm getAlgorithm() {
		return Algorithm.WINDOW;
	}

}
