package org.processmining.socialnetwork.models.impl.graphstream;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.swing.JComponent;

import org.graphstream.graph.Edge;
import org.graphstream.graph.Node;
import org.graphstream.graph.implementations.SingleGraph;
import org.graphstream.ui.spriteManager.Sprite;
import org.graphstream.ui.spriteManager.SpriteManager;
import org.processmining.framework.util.Pair;
import org.processmining.socialnetwork.models.SocialNetworkArcLabelStyle;
import org.processmining.socialnetwork.models.SocialNetworkArcStyle;
import org.processmining.socialnetwork.models.SocialNetworkVertexStyle;
import org.processmining.socialnetwork.models.SocialNetworkVisualization;

public class GraphStreamSingleGraphSocialNetworkVisualizationImpl<T> extends SingleGraph
		implements SocialNetworkVisualization<T> {

	static {
		// use a more advanced renderer for the graph
		System.setProperty("org.graphstream.ui.renderer", "org.graphstream.ui.j2dviewer.J2DGraphRenderer");
	}

	private final Map<Pair<T, T>, Edge> arcs = new HashMap<>();

	private final GraphStreamJPanel graphPanel;
	private SpriteManager sman;
	private final Map<T, Node> vertices = new ConcurrentHashMap<>();

	public GraphStreamSingleGraphSocialNetworkVisualizationImpl() {
		super("graph_stream_social_network_created_at_" + System.currentTimeMillis());
		graphPanel = new GraphStreamJPanel(this);
		sman = new SpriteManager(this);
		addAttribute("ui.stylesheet", GSSingleGraphDefaultLayout.LAYOUT);
		addAttribute("ui.quality");
		addAttribute("ui.antialias");
	}

	public void addArc(T from, T to) {
		addOrGetEdge(from, to);
	}

	private Edge addOrGetEdge(T from, T to) {
		Node fromNode = addOrGetNode(from);
		Node toNode = addOrGetNode(to);
		Pair<T, T> pair = new Pair<>(from, to);
		Edge edge = null;
		if (arcs.containsKey(pair)) {
			edge = arcs.get(pair);
		} else {
			edge = addEdge(constructLinkId(from, to), fromNode, toNode, true);
			arcs.put(pair, edge);
			Sprite sprite = sman.addSprite("s_" + edge.getId());
			sprite.setPosition(0.5, 0, 0);
			sprite.attachToEdge(edge.getId());
		}
		return edge;
	}

	private Node addOrGetNode(T t) {
		Node node = null;
		if (vertices.containsKey(t)) {
			node = vertices.get(t);
		} else {
			node = addNode(t.toString());
			vertices.put(t, node);
		}
		return node;
	}

	public void addVertex(T vertex) {
		addOrGetNode(vertex);
	}

	public void clear() {
		super.clear();
		arcs.clear();
		vertices.clear();
		sman = new SpriteManager(this);
		addAttribute("ui.stylesheet", GSSingleGraphDefaultLayout.LAYOUT);
	}

	private String constructLinkId(T from, T to) {
		return from.toString() + "->" + to.toString();
	}

	public boolean containsArc(T from, T to) {
		return arcs.containsKey(new Pair<T, T>(from, to));
	}

	public JComponent get() {
		return graphPanel;
	}

	private String getCSSStyleClass(SocialNetworkArcLabelStyle style) {
		switch (style) {
			case DOWN :
				return GSSingleGraphDefaultLayout.CLASS_DOWN;
			case UP :
				return GSSingleGraphDefaultLayout.CLASS_UP;
			default :
				return null;
		}
	}

	private String getCSSStyleClass(SocialNetworkArcStyle style) {
		switch (style) {
			case HIGH :
				return GSSingleGraphDefaultLayout.CLASS_HIGH;
			case LOW :
				return GSSingleGraphDefaultLayout.CLASS_LOW;
			case NEUTRAL :
				return GSSingleGraphDefaultLayout.CLASS_NEUTRAL;
			case VERY_HIGH :
				return GSSingleGraphDefaultLayout.CLASS_VERY_HIGH;
			case VERY_LOW :
				return GSSingleGraphDefaultLayout.CLASS_VERY_LOW;
			case DELETED :
				return GSSingleGraphDefaultLayout.CLASS_DELETED;
			case NEW :
				return GSSingleGraphDefaultLayout.CLASS_NEW;
		}
		return GSSingleGraphDefaultLayout.CLASS_NEUTRAL;
	}

	public SocialNetworkVisualization.Library getLibrary() {
		return SocialNetworkVisualization.Library.GRAPH_STREAM;
	}

	private boolean isDisconnected(T vertex) {
		boolean result = true;
		for (Pair<T, T> arc : arcs.keySet()) {
			if (arc.getFirst().equals(vertex) || arc.getSecond().equals(vertex)) {
				result = false;
				break;
			}
		}
		return result;
	}

	public void removeArc(T from, T to) {
		if (vertices.containsKey(from) && vertices.containsKey(to)) {
			Pair<T, T> pair = new Pair<T, T>(from, to);
			Edge e = arcs.get(pair);
			if (e != null) {
				removeEdge(e);
				arcs.remove(pair);
				sman.removeSprite("s_" + e.getId());
			}
			if (isDisconnected(from)) {
				removeVertex(from);
			}
			if (isDisconnected(to)) {
				removeVertex(to);
			}
		}
	}

	public void removeVertex(T vertex) {
		removeVertex(vertex, isDisconnected(vertex));
	}

	private void removeVertex(T vertex, boolean isDisconnected) {
		// remove all connected edges
		if (vertices.containsKey(vertex)) {
			if (isDisconnected(vertex)) {
				removeNode(vertices.get(vertex));
				vertices.remove(vertex);
			} else {
				Collection<Pair<T, T>> toRemove = new HashSet<>();
				for (Pair<T, T> arc : arcs.keySet()) {
					if (arc.getFirst().equals(vertex) || arc.getSecond().equals(vertex)) {
						toRemove.add(arc);
						removeEdge(arcs.get(arc));
					}
				}
				for (Pair<T, T> arc : toRemove) {
					arcs.remove(arc);
				}
				removeNode(vertices.get(vertex));
				vertices.remove(vertex);
				for (T t : vertices.keySet()) {
					boolean conn = isDisconnected(t);
					if (conn) {
						removeVertex(t, conn);
					}
				}

			}
		}
	}

	public void setArcLabel(T from, T to, String label) {
		Edge edge = addOrGetEdge(from, to);
		Sprite s = sman.getSprite("s_" + edge.getId());
		s.addAttribute("ui.label", label);
	}

	public void setArcLabelStyle(T from, T to, SocialNetworkArcLabelStyle style) {
		Sprite s = sman.getSprite("s_" + constructLinkId(from, to));
		s.changeAttribute("ui.class", getCSSStyleClass(style));
	}

	public void setArcStyle(T from, T to, SocialNetworkArcStyle style) {
		Edge e = addOrGetEdge(from, to);
		e.addAttribute("ui.class", getCSSStyleClass(style));
	}

	public void setArcStyle(T from, T to, SocialNetworkArcStyle... styles) {
		Edge e = addOrGetEdge(from, to);
		String stylesStr = "";
		for (int i = 0; i < styles.length; i++) {
			stylesStr += getCSSStyleClass(styles[i]);
			if (i < styles.length - 1) {
				stylesStr += ", ";
			}
		}
		e.addAttribute("ui.class", stylesStr);
	}

	public void setVertexLabel(T node, String label) {
		Node n = vertices.get(node);
		if (n != null) {
			n.addAttribute("ui.label", label);
		}
	}

	public void setVertexStyle(T node, SocialNetworkVertexStyle style) {
		if (vertices.containsKey(node)) {
			Node nodeInGraph = vertices.get(node);
			switch (style) {
				case HIGH :
					nodeInGraph.addAttribute("ui.class", GSSingleGraphDefaultLayout.CLASS_HIGH);
					break;
				case LOW :
					nodeInGraph.addAttribute("ui.class", GSSingleGraphDefaultLayout.CLASS_LOW);
					break;
				case NEUTRAL :
					nodeInGraph.addAttribute("ui.class", GSSingleGraphDefaultLayout.CLASS_NEUTRAL);
					break;
				case VERY_HIGH :
					nodeInGraph.addAttribute("ui.class", GSSingleGraphDefaultLayout.CLASS_VERY_HIGH);
					break;
				case VERY_LOW :
					nodeInGraph.addAttribute("ui.class", GSSingleGraphDefaultLayout.CLASS_VERY_LOW);
					break;
				case NEW :
					nodeInGraph.addAttribute("ui.class", GSSingleGraphDefaultLayout.CLASS_NEW);
					break;
				case DELETED :
					nodeInGraph.addAttribute("ui.class", GSSingleGraphDefaultLayout.CLASS_DELETED);
					break;
				default :
					break;

			}
		}

	}

	public String getArcLabel(T from, T to) {
		Edge edge = addOrGetEdge(from, to);
		Sprite s = sman.getSprite("s_" + edge.getId());
		return s.getAttribute("ui.label");
	}

	public boolean containsVertex(T vertex) {
		return vertices.containsKey(vertex);
	}

}
